/* 
 * Copyright (C) 1999 Robert Wilhelm
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <gnome.h>

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "makros.h"
#include "debug.h"
#include "position.h"
#include "pro.h"
#include "movlist.h"
#include "cho1.h"
#include "board.h"
#include "pgn.h"

#define DEBUG 

struct  pgn_tag
{
  char *event;
  char *site;
  char *date;
  char *round;
  char *white;
  char *black;
  char *result;
};

enum {
  PGN_EVENT,
  PGN_SITE,
  PGN_DATE,
  PGN_ROUND,
  PGN_WHITE,
  PGN_BLACK,
  PGN_RESULT, 
  PGN_WHITE_COUNTRY, 
  PGN_BLACK_COUNTRY, 
  PGN_WHITE_TITLE,
  PGN_BLACK_TITLE,
  PGN_WHITE_ELO,
  PGN_BLACK_ELO,
  PGN_REMARK,
  PGN_ECO,
  PGN_OPENING,
  PGN_VARIATION,
  PGN_PLYCOUNT,
  PGN_ANNOTATOR,
  PGN_TIMECONTROL,
  PGN_NIC,
  PGN_EVENTDATE
};

static struct
{
  gchar *name;
  gint token;
} symbols[] = {
  { "Event", PGN_EVENT },         
  { "Site", PGN_SITE },
  { "Date", PGN_DATE },
  { "Round", PGN_ROUND },
  { "White", PGN_WHITE },
  { "Black", PGN_BLACK },
  { "Result", PGN_RESULT },
  { "WhiteCountry", PGN_WHITE_COUNTRY },
  { "BlackCountry", PGN_BLACK_COUNTRY },
  { "WhiteTitle", PGN_WHITE_TITLE },
  { "BlackTitle", PGN_BLACK_TITLE },
  { "WhiteELO", PGN_WHITE_ELO },
  { "BlackELO", PGN_BLACK_ELO },
  { "ELOWhite", PGN_WHITE_ELO }, 	/* some broken programs use this
  { "ELOBlack", PGN_BLACK_ELO },	and this */
  { "Remark", PGN_REMARK },
  { "ECO", PGN_ECO },
  { "NIC", PGN_NIC },
  { "Opening", PGN_OPENING },
  { "Variation", PGN_VARIATION },
  { "Plycount", PGN_PLYCOUNT },
  { "Annotator", PGN_ANNOTATOR },
  { "EventDate", PGN_EVENTDATE },
  { "TimeControl", PGN_TIMECONTROL },
};

static guint nsymbols = sizeof (symbols) / sizeof (symbols[0]);

static GScannerConfig scanner_config =
{
  (
   " \t\n\r"
   )			/* cset_skip_characters */,
  (
   G_CSET_a_2_z
   "_"
   G_CSET_A_2_Z
   "_0123456789"
   )			/* cset_identifier_first */,
  (
   G_CSET_a_2_z
   "_0123456789"
   G_CSET_A_2_Z
   G_CSET_LATINS
   G_CSET_LATINC
   "-+\\='./;:,$"
   )			/* cset_identifier_nth */,
  ( ";\n" )		/* cpair_comment_single */,
  
  FALSE			/* case_sensitive */,
  
  TRUE			/* skip_comment_multi */,
  TRUE			/* skip_comment_single */,
  TRUE			/* scan_comment_multi */,
  TRUE			/* scan_identifier */,
  FALSE			/* scan_identifier_1char */,
  FALSE			/* scan_identifier_NULL */,
  TRUE			/* scan_symbols */,
  FALSE			/* scan_binary */,
  TRUE			/* scan_octal */,
  TRUE			/* scan_float */,
  TRUE			/* scan_hex */,
  FALSE			/* scan_hex_dollar */,
  TRUE			/* scan_string_sq */,
  TRUE			/* scan_string_dq */,
  TRUE			/* numbers_2_int */,
  FALSE			/* int_2_float */,
  FALSE			/* identifier_2_string */,
  TRUE			/* char_2_token */,
  FALSE			/* symbol_2_token */,
  FALSE			/* scope_0_fallback */
};

static char *filename;

static int txt_reg_move(char *str, int test);
static void pgn_header_load(GScanner *s, struct pgn_tag *);
static GScanner* pgn_scanner_init(void);
static GList * pgn_index(const char *fname);
static void  pgn_clist_add(GtkWidget * list ,struct pgn_tag tag);
static  GtkWidget  * pgn_clist_new(GList *game_list);
static void select_game(GtkWidget * widget,int row,int column,
                         GdkEventButton * bevent, GList *data);
static int txt_load_from_fp(FILE *fp, int len, int test);
static int txt_load(char *newbuf, int len, int test);
static void clist_click_column (GtkCList * clist, gint column, gpointer data);

#ifdef DEBUG
void pgn_test(void);
void pgn_test(void)
{
  int i,n;
  FILE *file;
  char *fname="ttt";
  GList *game_list;
  GList *game;

  filename = g_strdup(fname);
  game_list=pgn_index(fname);

  n = g_list_length(game_list); 
  file = fopen(fname, "r");

  game = game_list;

  for(i=0;i<n-1;i++)
    {
      fseek(file,(long)game->data,SEEK_SET);
      txt_load_from_fp(file,(long)game->next->data - (long)game->data, TRUE);
      game = game->next;
    }
}
#endif

/* HACK ALERT
   gscanner now has buffers and does no longer like external seeks.
   this hack works around the problem
   According to Tim Janik on gtk-devel (12 Nov 1998) 
   "this will hopefully change, as soon as an IO abstraction for GLib is
   implemented which can be used as an input to GScanner."          
*/


static void
g_scanner_sync_file_offset_extern (GScanner *scanner, gint offset )
{
  g_return_if_fail (scanner != NULL);

  if (scanner->input_fd >= 0)
    {
      if (lseek (scanner->input_fd, offset, SEEK_SET) >= 0)
	{
	  /* we succeeded, blow our buffer's contents now */
	  scanner->text = NULL;
	  scanner->text_end = NULL;
	}
    }
}

/* End of HACK ALERT */

void pgn_open(const char *fname)
{
  GtkWidget *clist;
  int i,n;
  int file;
  gchar *text;
  GList *game_list;
  GScanner *s = pgn_scanner_init();

  filename = g_strdup(fname);
  game_list=pgn_index(fname);
  n=g_list_length(game_list);
  clist = pgn_clist_new(game_list);

  file = open(fname, O_RDONLY);
  g_scanner_input_file(s,file);

  for(i=0;i<n-1;i++)
    {
      struct pgn_tag tag;
      
      g_scanner_sync_file_offset_extern(s,(long)g_list_nth_data(game_list,i));
      pgn_header_load(s, &tag);
      pgn_clist_add(clist,tag);
    }

	/* If there are any games set up the names and results properly */
	if (n>0) {
		gtk_clist_get_text(GTK_CLIST(clist), 0, 0, &text );
		gtk_label_set_text(GTK_LABEL(board.whitename), text);
		gtk_clist_get_text(GTK_CLIST(clist), 0, 1, &text );
		gtk_label_set_text(GTK_LABEL(board.blackname), text);
		gtk_clist_get_text(GTK_CLIST(clist), 0, 2, &text );
		gnomechess_movelist_set_result(board.movelist, text);
	}

  close(file);
  g_scanner_destroy(s);
}
  
static void  pgn_clist_add(GtkWidget * clist ,struct pgn_tag tag)
{

  char    *text[3] = { "", "","" };
  gint row;

  debug_print(DEBUG_NORMAL,"%s %s %s\n",tag.white,tag.black,tag.result);

  row=gtk_clist_append (GTK_CLIST(clist), text);
  gtk_clist_set_text (GTK_CLIST(clist), row , 0, tag.white);
  gtk_clist_set_text (GTK_CLIST(clist), row , 1, tag.black);
  gtk_clist_set_text (GTK_CLIST(clist), row , 2, tag.result);
  
}

static GList * pgn_index(const char *fname)
{
  long tell;

  FILE *fp = fopen(fname,"r");
  char buf[1024];
  char *p;

  int i=0;

  GList *game_list = NULL;

  if (!fp) return FALSE;

  while(1)
    {
      tell= ftell(fp);
      if ( !fgets(buf,sizeof(buf),fp)) break;

      if ((p=strstr(buf,"[Event \"")))
	{
	  i++;

	  /* for now use g_list_prepend(), because it is much faster 
		than g_list_append(), ( O(1) vs. O(n) ) */
	  game_list = g_list_prepend(game_list,(void *) (tell + (p-buf)));
	}
    }

  game_list = g_list_prepend(game_list,(void *) tell);
  game_list = g_list_reverse(game_list);
  fclose(fp);
  return game_list;
}

static int txt_load_from_fp(FILE *fp, int len, int test)
{
  int ret;

  char *buf = malloc(len);

  fread(buf,len,1,fp);
  ret = txt_load(buf,len,test);
  free(buf);
  return ret;
}

static GScanner* pgn_scanner_init()
{
  int i;
  GScanner *s = g_scanner_new(&scanner_config);

  for (i = 0; i < nsymbols; i++)
    g_scanner_add_symbol (s, symbols[i].name, 
			  GINT_TO_POINTER (symbols[i].token));
  return s;

}

static void pgntag(GTokenValue value1, GTokenValue value2, struct pgn_tag *tag)
{
	  switch (value1.v_int)
	    {
	    case PGN_EVENT:
	      tag->event = g_strdup(value2.v_string);
	      break;
	    case PGN_SITE:
	      tag->site = g_strdup(value2.v_string);
	      break;
	    case PGN_DATE:
	      tag->date = g_strdup(value2.v_string);
	      break;
	    case PGN_ROUND:
	      tag->round = g_strdup(value2.v_string);
	      break;
	    case PGN_WHITE:
	      tag->white = g_strdup(value2.v_string);
	      break;
	    case PGN_BLACK:
	      tag->black = g_strdup(value2.v_string);
	      break;
	    case PGN_RESULT:
	      tag->result = g_strdup(value2.v_string);
	      break;
	    default:
	      debug_print(DEBUG_NORMAL, "Unused PGN tag found: %s.\n",value2.v_string);
	    }
}

static struct pgn_tag * pgn_tag_new(void)
{
  struct pgn_tag *tag= malloc(sizeof( struct pgn_tag));

  tag->event=tag->site=tag->date=tag->round=tag->white=tag->black=tag->result=0;

  return tag;
}
static void pgn_tag_destroy(struct pgn_tag *tag)
{
  if (tag->event) g_free(tag->event);
  if (tag->site)  g_free(tag->site);
  if (tag->date)  g_free(tag->date);
  if (tag->round) g_free(tag->round);
  if (tag->white) g_free(tag->white);
  if (tag->black) g_free(tag->black);
  if (tag->result)g_free(tag->result);

  g_free(tag);
}

static int txt_load(char *newbuf, int len, int test)
{
  struct pgn_tag *tag= pgn_tag_new();
  int ret =FALSE;
  GScanner *s = pgn_scanner_init();

  gnomechess_position_set_initial(currPositionPtr);
  gnomechess_movelist_clear(board.movelist);

  g_scanner_input_text(s,newbuf,len);

  while(1)
    {
      GTokenValue value1,value2;
      GTokenType token2;
      GTokenType token =g_scanner_get_next_token(s);
      debug_print(DEBUG_VERBOSE, "Token: %d ",token);

      if (token == G_TOKEN_EOF) break;


      switch(token)
	{
	case  G_TOKEN_LEFT_BRACE:
	  token2 =g_scanner_get_next_token(s);
	  value1 = g_scanner_cur_value(s);

	  if (token2 == G_TOKEN_SYMBOL)
	    {
	      g_scanner_get_next_token(s);
	      value2 = g_scanner_cur_value(s);
	      pgntag(value1,value2,tag);
	    }
	  else
	    {
	      g_scanner_get_next_token(s);
	      value2 = g_scanner_cur_value(s);
	    }
	  token2 =g_scanner_get_next_token(s);
	  if (token2 != G_TOKEN_RIGHT_BRACE)
	    {
	      debug_print(DEBUG_NONE, "Parse error!\n");
	      ret = FALSE;
	      goto finish;
	    }
	  break;
	case  G_TOKEN_RIGHT_BRACE:
	case  G_TOKEN_SYMBOL:  
	case  G_TOKEN_STRING:
	  debug_print(DEBUG_NONE,"Parse error!\n");
	  break;
	case  G_TOKEN_FLOAT:
	  break;
	case  G_TOKEN_IDENTIFIER:
	  value1 = g_scanner_cur_value(s);
	  if (  '0' <=  *value1.v_string &&  *value1.v_string <= '9')
	    break;
	  txt_reg_move(value1.v_string, test);
	  break;
	case G_TOKEN_LEFT_CURLY: /* comments are ignored for now */
	  {
	    do 
	      {
		token2 =g_scanner_get_next_token(s);
		if (token2 ==  G_TOKEN_IDENTIFIER) 
		  {
		    value1 = g_scanner_cur_value(s);
		  }
		else 
		  if (token2 ==  G_TOKEN_EOF) 
		    { 
		      debug_print(DEBUG_NONE, "EOF in comment!\n");
		      ret = FALSE;
		      goto finish;
		    }
	      }
	    while(token2 != G_TOKEN_RIGHT_CURLY);
	    break;
	  }
	case G_TOKEN_LEFT_PAREN: /* RAVs are ignored for now */
	  { 
	    int nest=1;
	    debug_print(DEBUG_VERBOSE, "Nest %d\n",nest);
	    while(1) 
	      {
		token2 =g_scanner_get_next_token(s);
		debug_print(DEBUG_VERBOSE, "Token: %d\n",token2);
		if (token2 == G_TOKEN_EOF)
		  {
		    debug_print(DEBUG_NONE,"EOF in RAV!\n");
		    ret = FALSE;
		    goto finish;
		  }
		else
		if (token2 == G_TOKEN_LEFT_PAREN) 
		  {
		    nest++;
	    	debug_print(DEBUG_VERBOSE, "Nest %d\n",nest);
		  }
		else
		if (token2 == G_TOKEN_RIGHT_PAREN) 
		  {
		    nest--;
	    	debug_print(DEBUG_VERBOSE, "Unnest %d\n",nest);
		    if (!nest) break;
		  }

	      }
	  }
	  break;
	default:
	  debug_print(DEBUG_NONE, "Unknown token: %d\n",token);
	  ret = FALSE;
	  goto finish;
	}
    }

  ret = TRUE;

 finish:
  pgn_tag_destroy(tag);
  g_scanner_destroy(s);
  return ret;
}

static void pgn_header_load(GScanner *s, struct pgn_tag * tp)
{

  while(1)
    {
      GTokenValue value1,value2;
      GTokenType token =g_scanner_get_next_token(s);
	  debug_print(DEBUG_VERBOSE, "Token: %d\n",token);

	if (token == G_TOKEN_EOF) break;


      switch(token)
	{
	case  G_TOKEN_LEFT_BRACE:
	  break;
	case  G_TOKEN_RIGHT_BRACE:
	  break;
	case  G_TOKEN_SYMBOL:  
	  value1 = g_scanner_cur_value(s);
          g_scanner_get_next_token(s);
	  value2 = g_scanner_cur_value(s);
	  pgntag(value1,value2,tp);
	  break;
	case  G_TOKEN_STRING:
	  break;
	case  G_TOKEN_FLOAT:
	  break;
	case  G_TOKEN_IDENTIFIER:
	  value1 = g_scanner_cur_value(s);
	  goto finish;
	  break;
	  
	default:
	  debug_print(DEBUG_NONE, "Unknown token: %d\n",token);
	  return;
	}
    }

 finish:
}


static void delete_x(char *str)
{
  char *p = strchr(str,'x');

  if (p)
      while(( *p = *(p+1)))
	p++;
}

static void delete_plus(char *str)
{
  char *p = strchr(str,'+');

  if (p)
      while(( *p = *(p+1)))
	p++;
}

static void delete_ep(char *str)
{
  char *p = strstr(str,"ep");

  if (p)
      while(( *p = *(p+2)))
	p++;
}

static void delete_equal(char *str)
{
  char *p = strstr(str,"=");

  if (p)
      while(( *p = *(p+1)))
	p++;
}

static inline int txt_reg_move_found(gchar *aq, int test)
{
	    int piece=currPositionPtr->square[*aq];
	    gnomechess_position_move(currPositionPtr,*aq,*(aq+1));
	    if (!test) gnomechess_movelist_add(board.movelist, piece,*aq,*(aq+1));
	    return TRUE;		
}

static int txt_reg_move(char *str, int test)
{
  gshort anz,anz_n,anz_s;
  gshort i;
  gchar *ap,*aq;
  gchar *p;
  gchar zugliste[AB_ZUGL];
  gchar liste[100][10];

  delete_x(str);
  delete_plus(str);
  delete_ep(str);
  delete_equal(str);

  ap =  zugliste + AB_ZUG_S;
  anz = legal_move(&ap,&anz_s,&anz_n);

  for ( aq = ap , i = 0 ; i < anz ; i++ , aq += 2 ) 
  {
    p=liste[i];
    piece_move_to_ascii(p,currPositionPtr->square[(gshort)*aq],*aq,*(aq+1));
    if ( *p == ' ')  /* pawn move */
      {
	p++;         /* e.g. e2e4 */
	if (!strcmp(p,str))
	  return txt_reg_move_found(aq,test);
	p[1]=p[2];  /* e.g. ed5 */
	p[2]=p[3];
	p[3]=p[4];
	p[4]=p[5];
	if (p[0] != p[1])   /* not e.g. bb3 */
	  if (!strcmp(p,str))
	    return txt_reg_move_found(aq,test);

	p++;	 /* e.g. d5 */
	if (!strcmp(p,str))
	  return txt_reg_move_found(aq,test);

      }
    else  /* no pawn move */
      {
	char  tmp;

	if (!strcmp(p,str))  /* e.g. Ng1f3 */
	  return txt_reg_move_found(aq,test);

	tmp =p[2];
	p[2]=p[3];         /* Ngf3 */
	p[3]=p[4];
	p[4]=p[5];

	if (!strcmp(p,str))
	  return txt_reg_move_found(aq,test);

	p[1]=tmp;       /* N1f3 */

	if (!strcmp(p,str))
	  return txt_reg_move_found(aq,test);

	p[1]=p[2];    /* Nf3 */
	p[2]=p[3];
	p[3]=p[4];

	if (!strcmp(p,str))
	  return txt_reg_move_found(aq,test);
      }
  }
  return FALSE;
}

static char *titles[] = { N_("White"), N_("Black"), N_("Result")};

static  GtkWidget  * pgn_clist_new(GList *game_list)
{
  GtkWidget       *window;
  GtkWidget       *scrolled_window;
  GtkWidget       *clist;
  int i;

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  scrolled_window = gtk_scrolled_window_new (NULL, NULL);

  clist = gtk_clist_new_with_titles(3,titles);
  gtk_clist_set_selection_mode(GTK_CLIST (clist),GTK_SELECTION_BROWSE);

  for (i = 0; i < 3; i++)
      gtk_clist_set_column_auto_resize (GTK_CLIST (clist), i, TRUE);

  gtk_widget_set_usize (clist, 200, 200);

  gtk_widget_show (clist);
  gtk_container_add (GTK_CONTAINER (window), scrolled_window);
  gtk_container_add(GTK_CONTAINER(scrolled_window), clist);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),         
				  GTK_POLICY_AUTOMATIC,             
				  GTK_POLICY_AUTOMATIC);                  
  
  GTK_CLIST(clist)->sort_type = GTK_SORT_DESCENDING;

  gtk_signal_connect (GTK_OBJECT (clist), "click_column",
		      GTK_SIGNAL_FUNC (clist_click_column),
		      NULL);

  gtk_signal_connect (GTK_OBJECT (clist),
		      "select_row",
		      (GtkSignalFunc) select_game,
		      game_list);

  gtk_widget_show (scrolled_window);
  gtk_widget_show (window);
  return clist;
}

static void
clist_click_column (GtkCList * clist, gint column, gpointer data)
{
  if (column == clist->sort_column)
    {
      if (clist->sort_type == GTK_SORT_ASCENDING)
	clist->sort_type = GTK_SORT_DESCENDING;
      else
	clist->sort_type = GTK_SORT_ASCENDING;
    }
  else
    gtk_clist_set_sort_column (clist, column);

  gtk_clist_set_compare_func (clist, NULL);
  gtk_clist_sort (clist);
}

static void
select_game(GtkWidget * widget,int row,int column,GdkEventButton * event,GList *game_list)
{
	FILE *fp;
	gchar *text;

	fp = fopen(filename,"r");
	if (fp) {
		fseek(fp,(long) g_list_nth_data(game_list,row),SEEK_SET);
		txt_load_from_fp(fp, (long)g_list_nth_data(game_list,row+1) -
								(long)g_list_nth_data(game_list,row), FALSE);

		/* Now set up the names and results properly */
		gtk_clist_get_text(GTK_CLIST(widget), row, 0, &text );
		gtk_label_set_text(GTK_LABEL(board.whitename), text);
		gtk_clist_get_text(GTK_CLIST(widget), row, 1, &text );
		gtk_label_set_text(GTK_LABEL(board.blackname), text);
		gtk_clist_get_text(GTK_CLIST(widget), row, 2, &text );
		gnomechess_movelist_set_result(board.movelist, text);

		gnomechess_board_update();
    }
}

